import warnings
import numpy as np
import scipy.optimize as spopt
from scipy.constants import hbar
from scipy.interpolate import splrep, splev

from .utilities import plotting, save_load, Watt2dBm, dBm2Watt
from .circlefit import circlefit
from .calibration import calibration


##
## z_data_raw denotes the raw data
## z_data denotes the normalized data
##

class reflection_port(circlefit, save_load, plotting, calibration):
    '''
    normal direct port probed in reflection
    '''

    def __init__(self, f_data=None, z_data_raw=None):
        self.porttype = 'direct'
        self.fitresults = {}
        self.z_data = None
        if f_data is not None:
            self.f_data = np.array(f_data)
        else:
            self.f_data = None
        if z_data_raw is not None:
            self.z_data_raw = np.array(z_data_raw)
        else:
            self.z_data = None
        self.phasefitsmooth = 3

    def _S11(self, f, fr, k_c, k_i):
        '''
        use either frequency or angular frequency units
        for all quantities
        k_l=k_c+k_i: total (loaded) coupling rate
        k_c: coupling rate
        k_i: internal loss rate
        '''
        return ((k_c - k_i) + 2j * (f - fr)) / ((k_c + k_i) - 2j * (f - fr))

    def get_delay(self, f_data, z_data, delay=None, ignoreslope=True, guess=True):
        '''
        ignoreslope option not used here
        retrieves the cable delay assuming the ideal resonance has a circular shape
        modifies the cable delay until the shape Im(S21) vs Re(S21) is circular
        see "do_calibration"
        '''
        maxval = np.max(np.absolute(z_data))
        z_data = z_data / maxval
        A1, A2, A3, A4, fr, Ql = self._fit_skewed_lorentzian(f_data, z_data)
        if self.df_error / fr > 0.0001 or self.dQl_error / Ql > 0.1:
            # print "WARNING: Calibration using Lorentz fit failed, trying phase fit..."
            A1 = np.mean(np.absolute(z_data))
            A2 = 0.
            A3 = 0.
            A4 = 0.
            # fr = np.mean(f_data)
            f = splrep(f_data, np.unwrap(np.angle(z_data)), k=5, s=self.phasefitsmooth)
            fr = f_data[np.argmax(np.absolute(splev(f_data, f, der=1)))]
            Ql = 1e4
        if ignoreslope == True:
            A2 = 0.
        else:
            A2 = 0.
            # z_data = (np.absolute(z_data)-A2*(f_data-fr)) * np.exp(np.angle(z_data)*1j)  #usually not necessary
        if delay is None:
            if guess == True:
                delay = self._guess_delay(f_data, z_data)
            else:
                delay = 0.
            delay = self._fit_delay(f_data, z_data, delay, maxiter=200)
        params = [A1, A2, A3, A4, fr, Ql]
        return delay, params

    def do_calibration(self, f_data, z_data, ignoreslope=True, guessdelay=True, fixed_delay=None):
        '''
        calculating parameters for normalization
        '''
        delay, params = self.get_delay(f_data, z_data, ignoreslope=ignoreslope, guess=guessdelay,
                                       delay=fixed_delay)
        z_data = (z_data - params[1] * (f_data - params[4])) * np.exp(
            2. * 1j * np.pi * delay * f_data)
        xc, yc, r0 = self._fit_circle(z_data)
        zc = np.complex(xc, yc)
        fitparams = self._phase_fit(f_data, self._center(z_data, zc), 0., np.absolute(params[5]),
                                    params[4])
        theta, Ql, fr = fitparams
        beta = self._periodic_boundary(theta + np.pi, np.pi)  ###
        offrespoint = np.complex((xc + r0 * np.cos(beta)), (yc + r0 * np.sin(beta)))
        alpha = self._periodic_boundary(np.angle(offrespoint) + np.pi, np.pi)
        # a = np.absolute(offrespoint)
        # alpha = np.angle(zc)
        a = r0 + np.absolute(zc)
        return delay, a, alpha, fr, Ql, params[1], params[4]

    def do_normalization(self, f_data, z_data, delay, amp_norm, alpha, A2, frcal):
        '''
        transforming resonator into canonical position
        '''
        return (z_data - A2 * (f_data - frcal)) / amp_norm * np.exp(
            1j * (-alpha + 2. * np.pi * delay * f_data))

    def circlefit(self, f_data, z_data, fr=None, Ql=None, refine_results=False, calc_errors=True):
        '''
        S11 version of the circlefit
        '''

        if fr is None: fr = f_data[np.argmin(np.absolute(z_data))]
        if Ql is None: Ql = 1e6
        xc, yc, r0 = self._fit_circle(z_data, refine_results=refine_results)
        phi0 = -np.arcsin(yc / r0)
        theta0 = self._periodic_boundary(phi0 + np.pi, np.pi)
        z_data_corr = self._center(z_data, np.complex(xc, yc))
        theta0, Ql, fr = self._phase_fit(f_data, z_data_corr, theta0, Ql, fr)
        # print "Ql from phasefit is: " + str(Ql)
        Qi = Ql / (1. - r0)
        Qc = 1. / (1. / Ql - 1. / Qi)

        results = {"Qi": Qi, "Qc": Qc, "Ql": Ql, "fr": fr, "theta0": theta0}

        # calculation of the error
        p = [fr, Qc, Ql]
        # chi_square, errors = rt.get_errors(rt.residuals_notch_ideal,f_data,z_data,p)
        if calc_errors == True:
            chi_square, cov = self._get_cov_fast_directrefl(f_data, z_data, p)
            # chi_square, cov = rt.get_cov(rt.residuals_notch_ideal,f_data,z_data,p)

            if cov is not None:
                errors = np.sqrt(np.diagonal(cov))
                fr_err, Qc_err, Ql_err = errors
                # calc Qi with error prop (sum the squares of the variances and covariaces)
                dQl = 1. / ((1. / Ql - 1. / Qc) ** 2 * Ql ** 2)
                dQc = - 1. / ((1. / Ql - 1. / Qc) ** 2 * Qc ** 2)
                Qi_err = np.sqrt((dQl ** 2 * cov[2][2]) + (dQc ** 2 * cov[1][1]) + (
                        2 * dQl * dQc * cov[2][1]))  # with correlations
                errors = {"Ql_err": Ql_err, "Qc_err": Qc_err, "fr_err": fr_err,
                          "chi_square": chi_square, "Qi_err": Qi_err}
                results.update(errors)
            else:
                print("WARNING: Error calculation failed!")
        else:
            # just calc chisquared:
            fun2 = lambda x: self._residuals_notch_ideal(x, f_data, z_data) ** 2
            chi_square = 1. / float(len(f_data) - len(p)) * (fun2(p)).sum()
            errors = {"chi_square": chi_square}
            results.update(errors)

        return results

    def autofit(self, electric_delay=None):
        '''
        automatic calibration and fitting
        '''
        delay, amp_norm, alpha, fr, Ql, A2, frcal = \
            self.do_calibration(self.f_data, self.z_data_raw, ignoreslope=True, guessdelay=False,
                                fixed_delay=electric_delay)
        self.z_data = self.do_normalization(self.f_data, self.z_data_raw, delay, amp_norm, alpha,
                                            A2, frcal)
        self.fitresults = self.circlefit(self.f_data, self.z_data, fr, Ql, refine_results=False,
                                         calc_errors=True)
        self.z_data_sim = A2 * (self.f_data - frcal) + self._S11_directrefl(self.f_data,
                                                                            fr=self.fitresults[
                                                                                "fr"],
                                                                            Ql=self.fitresults[
                                                                                "Ql"],
                                                                            Qc=self.fitresults[
                                                                                "Qc"], a=amp_norm,
                                                                            alpha=alpha,
                                                                            delay=delay)

    def _S11_directrefl(self, f, fr=10e9, Ql=900, Qc=1000., a=1., alpha=0., delay=.0):
        '''
        full model for notch type resonances
        '''
        return a * np.exp(np.complex(0, alpha)) * np.exp(-2j * np.pi * f * delay) * (
                2. * Ql / Qc - 1. + 2j * Ql * (fr - f) / fr) / (1. - 2j * Ql * (fr - f) / fr)

    def get_single_photon_limit(self, unit='dBm'):
        '''
        returns the amout of power in units of W necessary
        to maintain one photon on average in the cavity
        unit can be 'dbm' or 'watt'
        '''
        if self.fitresults != {}:
            fr = self.fitresults['fr']
            k_c = 2 * np.pi * fr / self.fitresults['Qc']
            k_i = 2 * np.pi * fr / self.fitresults['Qi']
            if unit == 'dBm':
                return Watt2dBm(1. / (4. * k_c / (2. * np.pi * hbar * fr * (k_c + k_i) ** 2)))
            elif unit == 'watt':
                return 1. / (4. * k_c / (2. * np.pi * hbar * fr * (k_c + k_i) ** 2))

        else:
            warnings.warn('Please perform the fit first', UserWarning)
            return None

    def get_photons_in_resonator(self, power, unit='dBm'):
        '''
        returns the average number of photons
        for a given power (defaul unit is 'dbm')
        unit can be 'dBm' or 'watt'
        '''
        if self.fitresults != {}:
            if unit == 'dBm':
                power = dBm2Watt(power)
            fr = self.fitresults['fr']
            k_c = 2 * np.pi * fr / self.fitresults['Qc']
            k_i = 2 * np.pi * fr / self.fitresults['Qi']
            return 4. * k_c / (2. * np.pi * hbar * fr * (k_c + k_i) ** 2) * power
        else:
            warnings.warn('Please perform the fit first', UserWarning)
            return None


class notch_port(circlefit, save_load, plotting, calibration):
    '''
    notch type port probed in transmission
    '''

    def __init__(self, f_data=None, z_data_raw=None):
        self.porttype = 'notch'
        self.fitresults = {}
        self.z_data = None
        if f_data is not None:
            self.f_data = np.array(f_data)
        else:
            self.f_data = None
        if z_data_raw is not None:
            self.z_data_raw = np.array(z_data_raw)
        else:
            self.z_data_raw = None

    def get_delay(self, f_data, z_data, delay=None, ignoreslope=True, guess=True):
        '''
        retrieves the cable delay assuming the ideal resonance has a circular shape
        modifies the cable delay until the shape Im(S21) vs Re(S21) is circular
        see "do_calibration"
        '''
        maxval = np.max(np.absolute(z_data))
        z_data = z_data / maxval
        A1, A2, A3, A4, fr, Ql = self._fit_skewed_lorentzian(f_data, z_data)
        if ignoreslope == True:
            A2 = 0.
        else:
            A2 = 0.
            # z_data = (np.absolute(z_data)-A2*(f_data-fr)) * np.exp(np.angle(z_data)*1j)  #usually not necessary
        if delay is None:
            if guess == True:
                delay = self._guess_delay(f_data, z_data)
            else:
                delay = 0.
            delay = self._fit_delay(f_data, z_data, delay, maxiter=200)
        params = [A1, A2, A3, A4, fr, Ql]
        return delay, params

    def do_calibration(self, f_data, z_data, ignoreslope=True, guessdelay=True):
        '''
        performs an automated calibration and tries to determine the prefactors a, alpha, delay
        fr, Ql, and a possible slope are extra information, which can be used as start parameters for subsequent fits
        see also "do_normalization"
        the calibration procedure works for transmission line resonators as well
        '''
        delay, params = self.get_delay(f_data, z_data, ignoreslope=ignoreslope, guess=guessdelay)
        z_data = (z_data - params[1] * (f_data - params[4])) * np.exp(
            2. * 1j * np.pi * delay * f_data)
        xc, yc, r0 = self._fit_circle(z_data)
        zc = np.complex(xc, yc)
        fitparams = self._phase_fit(f_data, self._center(z_data, zc), 0., np.absolute(params[5]),
                                    params[4])
        theta, Ql, fr = fitparams
        beta = self._periodic_boundary(theta + np.pi, np.pi)
        offrespoint = np.complex((xc + r0 * np.cos(beta)), (yc + r0 * np.sin(beta)))
        alpha = np.angle(offrespoint)
        a = np.absolute(offrespoint)
        return delay, a, alpha, fr, Ql, params[1], params[4]

    def do_normalization(self, f_data, z_data, delay, amp_norm, alpha, A2, frcal):
        '''
        removes the prefactors a, alpha, delay and returns the calibrated data, see also "do_calibration"
        works also for transmission line resonators
        '''
        return (z_data - A2 * (f_data - frcal)) / amp_norm * np.exp(
            1j * (-alpha + 2. * np.pi * delay * f_data))

    def circlefit(self, f_data, z_data, fr=None, Ql=None, refine_results=False, calc_errors=True):
        '''
        performs a circle fit on a frequency vs. complex resonator scattering data set
        Data has to be normalized!!
        INPUT:
        f_data,z_data: input data (frequency, complex S21 data)
        OUTPUT:
        outpus a dictionary {key:value} consisting of the fit values, errors and status information about the fit
        values: {"phi0":phi0, "Ql":Ql, "absolute(Qc)":absQc, "Qi": Qi, "electronic_delay":delay, "complexQc":complQc, "resonance_freq":fr, "prefactor_a":a, "prefactor_alpha":alpha}
        errors: {"phi0_err":phi0_err, "Ql_err":Ql_err, "absolute(Qc)_err":absQc_err, "Qi_err": Qi_err, "electronic_delay_err":delay_err, "resonance_freq_err":fr_err, "prefactor_a_err":a_err, "prefactor_alpha_err":alpha_err}
        for details, see:
            [1] (not diameter corrected) Jiansong Gao, "The Physics of Superconducting Microwave Resonators" (PhD Thesis), Appendix E, California Institute of Technology, (2008)
            [2] (diameter corrected) M. S. Khalil, et. al., J. Appl. Phys. 111, 054510 (2012)
            [3] (fitting techniques) N. CHERNOV AND C. LESORT, "Least Squares Fitting of Circles", Journal of Mathematical Imaging and Vision 23, 239, (2005)
            [4] (further fitting techniques) P. J. Petersan, S. M. Anlage, J. Appl. Phys, 84, 3392 (1998)
        the program fits the circle with the algebraic technique described in [3], the rest of the fitting is done with the scipy.optimize least square fitting toolbox
        also, check out [5] S. Probst et al. "Efficient and reliable analysis of noisy complex scatterung resonator data for superconducting quantum circuits" (in preparation)
        '''

        if fr is None: fr = f_data[np.argmin(np.absolute(z_data))]
        if Ql is None: Ql = 1e6
        xc, yc, r0 = self._fit_circle(z_data, refine_results=refine_results)
        phi0 = -np.arcsin(yc / r0)
        theta0 = self._periodic_boundary(phi0 + np.pi, np.pi)
        z_data_corr = self._center(z_data, np.complex(xc, yc))
        theta0, Ql, fr = self._phase_fit(f_data, z_data_corr, theta0, Ql, fr)
        # print "Ql from phasefit is: " + str(Ql)
        absQc = Ql / (2. * r0)
        complQc = absQc * np.exp(1j * ((-1.) * phi0))
        Qc = 1. / (
                1. / complQc).real  # here, taking the real part of (1/complQc) from diameter correction method
        Qi_dia_corr = 1. / (1. / Ql - 1. / Qc)
        Qi_no_corr = 1. / (1. / Ql - 1. / absQc)

        results = {"Qi_dia_corr": Qi_dia_corr, "Qi_no_corr": Qi_no_corr, "absQc": absQc,
                   "Qc_dia_corr": Qc, "Ql": Ql, "fr": fr, "theta0": theta0, "phi0": phi0}

        # calculation of the error
        p = [fr, absQc, Ql, phi0]
        # chi_square, errors = rt.get_errors(rt.residuals_notch_ideal,f_data,z_data,p)
        if calc_errors == True:
            chi_square, cov = self._get_cov_fast_notch(f_data, z_data, p)
            # chi_square, cov = rt.get_cov(rt.residuals_notch_ideal,f_data,z_data,p)

            if cov is not None:
                errors = np.sqrt(np.diagonal(cov))
                fr_err, absQc_err, Ql_err, phi0_err = errors
                # calc Qi with error prop (sum the squares of the variances and covariaces)
                dQl = 1. / ((1. / Ql - 1. / absQc) ** 2 * Ql ** 2)
                dabsQc = - 1. / ((1. / Ql - 1. / absQc) ** 2 * absQc ** 2)
                Qi_no_corr_err = np.sqrt((dQl ** 2 * cov[2][2]) + (dabsQc ** 2 * cov[1][1]) + (
                        2 * dQl * dabsQc * cov[2][1]))  # with correlations
                # calc Qi dia corr with error prop
                dQl = 1 / ((1 / Ql - np.cos(phi0) / absQc) ** 2 * Ql ** 2)
                dabsQc = -np.cos(phi0) / ((1 / Ql - np.cos(phi0) / absQc) ** 2 * absQc ** 2)
                dphi0 = -np.sin(phi0) / ((1 / Ql - np.cos(phi0) / absQc) ** 2 * absQc)
                ##err1 = ( (dQl*cov[2][2])**2 + (dabsQc*cov[1][1])**2 + (dphi0*cov[3][3])**2 )
                err1 = ((dQl ** 2 * cov[2][2]) + (dabsQc ** 2 * cov[1][1]) + (
                        dphi0 ** 2 * cov[3][3]))
                err2 = (dQl * dabsQc * cov[2][1] + dQl * dphi0 * cov[2][3] + dabsQc * dphi0 *
                        cov[1][3])
                Qi_dia_corr_err = np.sqrt(err1 + 2 * err2)  # including correlations
                errors = {"phi0_err": phi0_err, "Ql_err": Ql_err, "absQc_err": absQc_err,
                          "fr_err": fr_err, "chi_square": chi_square,
                          "Qi_no_corr_err": Qi_no_corr_err, "Qi_dia_corr_err": Qi_dia_corr_err}
                results.update(errors)
            else:
                print("WARNING: Error calculation failed!")
        else:
            # just calc chisquared:
            fun2 = lambda x: self._residuals_notch_ideal(x, f_data, z_data) ** 2
            chi_square = 1. / float(len(f_data) - len(p)) * (fun2(p)).sum()
            errors = {"chi_square": chi_square}
            results.update(errors)

        return results

    def autofit(self, calc_errors = False):
        '''
        automatic calibration and fitting
        '''
        delay, amp_norm, alpha, fr, Ql, A2, frcal = \
            self.do_calibration(self.f_data, self.z_data_raw, ignoreslope=True, guessdelay=True)
        self.z_data = self.do_normalization(self.f_data, self.z_data_raw, delay, amp_norm, alpha,
                                            A2, frcal)
        self.fitresults = self.circlefit(self.f_data, self.z_data, fr, Ql, refine_results=False,
                                         calc_errors=calc_errors)
        self.fitresults["delay"] = delay
        self.fitresults["a"] = amp_norm
        self.fitresults["alpha"] = alpha
        self.z_data_sim = A2 * (self.f_data - frcal) + self._S21_notch(self.f_data,
                                                                       fr=self.fitresults["fr"],
                                                                       Ql=self.fitresults["Ql"],
                                                                       Qc=self.fitresults["absQc"],
                                                                       phi=self.fitresults["phi0"],
                                                                       a=amp_norm, alpha=alpha,
                                                                       delay=delay)

    def _S21_notch(self, f, fr=10e9, Ql=900, Qc=1000., phi=0., a=1., alpha=0., delay=.0):
        '''
        full model for notch type resonances
        '''
        return a * np.exp(np.complex(0, alpha)) * np.exp(-2j * np.pi * f * delay) * (
                1. - Ql / Qc * np.exp(1j * phi) / (1. + 2j * Ql * (f - fr) / fr))

    def get_single_photon_limit(self, unit='dBm', diacorr=True):
        '''
        returns the amout of power in units of W necessary
        to maintain one photon on average in the cavity
        unit can be 'dBm' or 'watt'
        '''
        if self.fitresults != {}:
            fr = self.fitresults['fr']
            if diacorr:
                k_c = 2 * np.pi * fr / self.fitresults['Qc_dia_corr']
                k_i = 2 * np.pi * fr / self.fitresults['Qi_dia_corr']
            else:
                k_c = 2 * np.pi * fr / self.fitresults['absQc']
                k_i = 2 * np.pi * fr / self.fitresults['Qi_no_corr']
            if unit == 'dBm':
                return Watt2dBm(1. / (4. * k_c / (2. * np.pi * hbar * fr * (k_c + k_i) ** 2)))
            elif unit == 'watt':
                return 1. / (4. * k_c / (2. * np.pi * hbar * fr * (k_c + k_i) ** 2))
        else:
            warnings.warn('Please perform the fit first', UserWarning)
            return None

    def get_photons_in_resonator(self, power, unit='dBm', diacorr=True):
        '''
        returns the average number of photons
        for a given power in units of W
        unit can be 'dBm' or 'watt'
        '''
        if self.fitresults != {}:
            if unit == 'dBm':
                power = dBm2Watt(power)
            fr = self.fitresults['fr']
            if diacorr:
                k_c = 2 * np.pi * fr / self.fitresults['Qc_dia_corr']
                k_i = 2 * np.pi * fr / self.fitresults['Qi_dia_corr']
            else:
                k_c = 2 * np.pi * fr / self.fitresults['absQc']
                k_i = 2 * np.pi * fr / self.fitresults['Qi_no_corr']
            return 4. * k_c / (2. * np.pi * hbar * fr * (k_c + k_i) ** 2) * power
        else:
            warnings.warn('Please perform the fit first', UserWarning)
            return None


class transmission_port(circlefit, save_load, plotting):
    '''
    a class for handling transmission measurements
    '''

    def __init__(self, f_data=None, z_data_raw=None):
        self.porttype = 'transm'
        self.fitresults = {}
        if f_data is not None:
            self.f_data = np.array(f_data)
        else:
            self.f_data = None
        if z_data_raw is not None:
            self.z_data_raw = np.array(z_data_raw)
        else:
            self.z_data = None

    def _S21(self, f, fr, Ql, A):
        return A ** 2 / (1. + 4. * Ql ** 2 * ((f - fr) / fr) ** 2)

    def fit(self):
        self.ampsqr = (np.absolute(self.z_data_raw)) ** 2
        p = [self.f_data[np.argmax(self.ampsqr)], 1000., np.amax(self.ampsqr)]
        popt, pcov = spopt.curve_fit(self._S21, self.f_data, self.ampsqr, p)
        errors = np.sqrt(np.diag(pcov))
        self.fitresults = {'fr': popt[0], 'fr_err': errors[0], 'Ql': popt[1], 'Ql_err': errors[1],
                           'Ampsqr': popt[2], 'Ampsqr_err': errors[2]}


class resonator(object):
    '''
    Universal resonator analysis class
    It can handle different kinds of ports and assymetric resonators.
    '''

    def __init__(self, ports={}, comment=None):
        '''
        initializes the resonator class object
        ports (dictionary {key:value}): specify the name and properties of the coupling ports
            e.g. ports = {'1':'direct', '2':'notch'}
        comment: add a comment
        '''
        self.comment = comment
        self.port = {}
        self.transm = {}
        if len(ports) > 0:
            for key, pname in ports.items():
                if pname == 'direct':
                    self.port.update({key: reflection_port()})
                elif pname == 'notch':
                    self.port.update({key: notch_port()})
                else:
                    warnings.warn("Undefined input type! Use 'direct' or 'notch'.", SyntaxWarning)
        if len(self.port) == 0: warnings.warn("Resonator has no coupling ports!", UserWarning)

    def add_port(self, key, pname):
        if pname == 'direct':
            self.port.update({key: reflection_port()})
        elif pname == 'notch':
            self.port.update({key: notch_port()})
        else:
            warnings.warn("Undefined input type! Use 'direct' or 'notch'.", SyntaxWarning)
        if len(self.port) == 0: warnings.warn("Resonator has no coupling ports!", UserWarning)

    def delete_port(self, key):
        del self.port[key]
        if len(self.port) == 0: warnings.warn("Resonator has no coupling ports!", UserWarning)

    def get_Qi(self):
        '''
        based on the number of ports and the corresponding measurements
        it calculates the internal losses
        '''
        pass

    def get_single_photon_limit(self, port):
        '''
        returns the amout of power necessary to maintain one photon
        on average in the cavity
        '''
        pass

    def get_photons_in_resonator(self, power, port):
        '''
        returns the average number of photons
        for a given power
        '''
        pass

    def add_transm_meas(self, port1, port2):
        '''
        input: port1
        output: port2
        adds a transmission measurement
        connecting two direct ports S21
        '''
        key = port1 + " -> " + port2
        self.port.update({key: transm()})
        pass


class batch_processing(object):
    '''
    A class for batch processing of resonator data as a function of another variable
    Typical applications are power scans, magnetic field scans etc.
    '''

    def __init__(self, porttype):
        '''
        porttype = 'notch', 'direct', 'transm'
        results is an array of dictionaries containing the fitresults
        '''
        self.porttype = porttype
        self.results = []

    def autofit(self, cal_dataslice=0):
        '''
        fits all data
        cal_dataslice: choose scatteringdata which should be used for calibration
        of the amplitude and phase, default = 0 (first)
        '''
        pass


class coupled_resonators(batch_processing):
    '''
    A class for fitting a resonator coupled to a second one
    '''

    def __init__(self, porttype):
        self.porttype = porttype
        self.results = []




######
## Functions to evaluate noise data
######

import warnings
import numpy as np
from scipy.signal import periodogram


class noisedata(object):
    
    def __init__(self,IQ,IQref,fr,Ql,fs,gain_corr=[1.,1.],Z=50):
        '''
        units are assumed to be in volts
        -> IQ = I+1j*Q ; with amplitude signal on Q and phase on I
        this signal is measured on resonance
        -> IQref = Iref+1j*Qref ; with amplitude signal on Qref and phase on Iref
        this signal is measured far off resonance
        IMPORTANT: IQ and IQref describe signals on opposite sides of the resonance circle
        Therefore, take care that Q and Qref have the correct signs in order that 
        the program can determine the diameter of the resonance circle.
        -> fr: resonance frequency
        -> Ql: loaded Q of the resonator
        -> fs: sampling rate
        -> gain_corr = [1.,1.] ; enter here if the gain of IQ and IQref signals
        are different
        -> Z: impedance
        The signals will be normalized to the reference such that IQref = 1.        
        '''
        self.Z = Z
        self.fr = fr
        self.Ql = Ql
        self.offrespoint = np.mean(np.imag(IQref))
        self.respoint = np.mean(np.imag(IQref))
        self.radius = (self.offrespoint - self.respoint)/self.offrespoint
        self.P_I = periodogram(self._demean(np.real(IQ)),fs=fs)
        self.P_Q = periodogram(self._demean(np.imag(IQ)),fs=fs)
        self.P_Iref = periodogram(self._demean(np.real(IQref)),fs=fs)
        self.P_Qref = periodogram(self._demean(np.imag(IQref)),fs=fs)

################################# 
        #functions to evalate multiple things
    def P_I_eval_all(self):
        '''
        returns a 2D numpy array with all the results
        and a 1D list with the description
        '''
        comment = ['P_I','P_Inorm','P_Ipower','P_dtheta','P_dphi','P_df','P_']
        return np.vstack((self.P_I,self.P_Inorm(),self.P_Ipower(),self.P_dtheta(),self.P_dphi(),self.P_df(),self.P_())), comment
        
    def P_Iref_eval_all(self):
        '''
        returns a 2D numpy array with all the results
        and a 1D list with the description
        '''
        comment = ['P_Iref','P_Irefnorm','P_Irefpower','P_refdtheta','P_refdphi','P_refdf','P_ref']
        return np.vstack((self.P_Iref,self.P_Irefnorm(),self.P_Irefpower(),self.P_refdtheta(),self.P_refdphi(),self.P_refdf(),self.P_ref())), comment

    def P_Q_eval_all(self):
        '''
        returns a 2D numpy array with all the results
        and a 1D list with the description
        '''
        comment = ['P_Q','P_Qnorm','P_Qpower']
        return np.vstack((self.P_Q,self.P_Qnorm(),self.P_Qpower())), comment
        
    def P_Qref_eval_all(self):
        '''
        returns a 2D numpy array with all the results
        and a 1D list with the description
        '''
        comment = ['P_Qref','P_Qrefnorm','P_Qrefpower']
        return np.vstack((self.P_Qref,self.P_Qrefnorm(),self.P_Qrefpower())), comment

################################# 
        #helpers
    def _demean(self,x):
        '''
        removes the mean value from x
        '''
        return x - x.mean()
    
        
################################# 
        #noise on I
        
    def P_Inorm(self):
        '''
        V^2/Hz
        '''
        return self.P_I/(self.offrespoint**2)
        
    def P_Ipower(self):
        '''
        W/Hz
        '''
        return self.P_I/self.Z
        
    def P_dtheta(self):
        '''
        rad^2/Hz
        phase noise on the resonator circle phase
        (this is not the real measured phase)
        '''
        return self.P_Inorm()/self.r**2
        
    def P_dphi(self):
        '''
        rad^2/Hz
        phase noise on the phase measured with the VNA
        '''
        return self.P_Inorm()/np.absolute(self.respoint**2)
        
    def P_df(self):
        '''
        Hz^2/Hz
        frequency noise
        '''
        return self.P_theta() * self.fr**2 / (16.*self.Ql**2)
        
    def P_(self):
        '''
        1/Hz
        fractional frequency noise
        '''
        return self.P_theta() / (16.*self.Ql**2)
        
#################################        
        #noise on Iref
        
    def P_Irefnorm(self):
        '''
        V^2/Hz
        '''
        return self.P_Iref/(self.offrespoint**2)
        
    def P_Irefpower(self):
        '''
        W/Hz
        '''
        return self.P_Iref/self.Z
        
    def P_refdtheta(self):
        '''
        rad^2/Hz
        phase noise on the resonator circle phase
        (this is not the real measured phase)
        '''
        return self.P_Irefnorm()/self.r**2
        
    def P_refdphi(self):
        '''
        rad^2/Hz
        phase noise on the phase measured with the VNA
        '''
        return self.P_Irefnorm()/np.absolute(self.respoint**2)
        
    def P_refdf(self):
        '''
        Hz^2/Hz
        frequency noise
        '''
        return self.P_reftheta() * self.fr**2 / (16.*self.Ql**2)
        
    def P_ref(self):
        '''
        1/Hz
        fractional frequency noise
        '''
        return self.P_reftheta() / (16.*self.Ql**2)
        
#################################        
        #noise on Q
        
    def P_Qnorm(self):
        '''
        V^2/Hz
        '''
        return self.P_Q/(self.offrespoint**2)
        
    def P_Qpower(self):
        '''
        W/Hz
        '''
        return self.P_Q/self.Z
        
#################################        
        #noise on Qref
        
    def P_Qrefnorm(self):
        '''
        V^2/Hz
        '''
        return self.P_Qref/(self.offrespoint**2)
        
    def P_Qrefpower(self):
        '''
        W/Hz
        '''
        return self.P_Qref/self.Z
        



import warnings
import numpy as np
import matplotlib.pyplot as plt

def Watt2dBm(x):
    '''
    converts from units of watts to dBm
    '''
    return 10.*np.log10(x*1000.)

def dBm2Watt(x):
    '''
    converts from units of watts to dBm
    '''
    return 10**(x/10.) /1000.

class plotting(object):
    '''
    some helper functions for plotting
    '''
    def plotall(self):
        real = self.z_data_raw.real
        imag = self.z_data_raw.imag
        real2 = self.z_data_sim.real
        imag2 = self.z_data_sim.imag
        fig = plt.figure(figsize=(15,5))
        fig.canvas.set_window_title("Resonator fit")
        plt.subplot(131)
        plt.plot(real,imag,label='rawdata')
        plt.plot(real2,imag2,label='fit')
        plt.xlabel('Re(S21)')
        plt.ylabel('Im(S21)')
        plt.legend()
        plt.subplot(132)
        plt.plot(self.f_data*1e-9,np.absolute(self.z_data_raw),label='rawdata')
        plt.plot(self.f_data*1e-9,np.absolute(self.z_data_sim),label='fit')
        plt.xlabel('f (GHz)')
        plt.ylabel('Amplitude')
        plt.legend()
        plt.subplot(133)
        plt.plot(self.f_data*1e-9,np.unwrap(np.angle(self.z_data_raw)),label='rawdata')
        plt.plot(self.f_data*1e-9,np.unwrap(np.angle(self.z_data_sim)),label='fit')
        plt.xlabel('f (GHz)')
        plt.ylabel('Phase')
        plt.legend()
        # plt.gcf().set_size_inches(15,5)
        plt.tight_layout()
        plt.show()


    def plotcalibrateddata(self):
        real = self.z_data.real
        imag = self.z_data.imag
        plt.subplot(221)
        plt.plot(real,imag,label='rawdata')
        plt.xlabel('Re(S21)')
        plt.ylabel('Im(S21)')
        plt.legend()
        plt.subplot(222)
        plt.plot(self.f_data*1e-9,np.absolute(self.z_data),label='rawdata')
        plt.xlabel('f (GHz)')
        plt.ylabel('|S21|')
        plt.legend()
        plt.subplot(223)
        plt.plot(self.f_data*1e-9,np.angle(self.z_data),label='rawdata')
        plt.xlabel('f (GHz)')
        plt.ylabel('arg(|S21|)')
        plt.legend()
        plt.show()

    def plotrawdata(self):
        real = self.z_data_raw.real
        imag = self.z_data_raw.imag
        plt.subplot(221)
        plt.plot(real,imag,label='rawdata')
        plt.xlabel('Re(S21)')
        plt.ylabel('Im(S21)')
        plt.legend()
        plt.subplot(222)
        plt.plot(self.f_data*1e-9,np.absolute(self.z_data_raw),label='rawdata')
        plt.xlabel('f (GHz)')
        plt.ylabel('|S21|')
        plt.legend()
        plt.subplot(223)
        plt.plot(self.f_data*1e-9,np.angle(self.z_data_raw),label='rawdata')
        plt.xlabel('f (GHz)')
        plt.ylabel('arg(|S21|)')
        plt.legend()
        plt.show()

class save_load(object):
    '''
    procedures for loading and saving data used by other classes
    '''
    def _ConvToCompl(self,x,y,dtype):
        '''
        dtype = 'realimag', 'dBmagphaserad', 'linmagphaserad', 'dBmagphasedeg', 'linmagphasedeg'
        '''
        if dtype=='realimag':
            return x+1j*y
        elif dtype=='linmagphaserad':
            return x*np.exp(1j*y)
        elif dtype=='dBmagphaserad':
            return 10**(x/20.)*np.exp(1j*y)
        elif dtype=='linmagphasedeg':
            return x*np.exp(1j*y/180.*np.pi)
        elif dtype=='dBmagphasedeg':
            return 10**(x/20.)*np.exp(1j*y/180.*np.pi)
        else: warnings.warn("Undefined input type! Use 'realimag', 'dBmagphaserad', 'linmagphaserad', 'dBmagphasedeg' or 'linmagphasedeg'.", SyntaxWarning)

    def add_data(self,f_data,z_data):
        self.f_data = np.array(f_data)
        self.z_data_raw = np.array(z_data)

    def cut_data(self,f1,f2):
        def findpos(f_data,val):
            pos = 0
            for i in range(len(f_data)):
                if f_data[i]<val: pos=i
            return pos
        pos1 = findpos(self.f_data,f1)
        pos2 = findpos(self.f_data,f2)
        self.f_data = self.f_data[pos1:pos2]
        self.z_data_raw = self.z_data_raw[pos1:pos2]

    def add_fromtxt(self,fname,dtype,header_rows,usecols=(0,1,2),fdata_unit=1.,delimiter=None):
        '''
        dtype = 'realimag', 'dBmagphaserad', 'linmagphaserad', 'dBmagphasedeg', 'linmagphasedeg'
        '''
        data = np.loadtxt(fname,usecols=usecols,skiprows=header_rows,delimiter=delimiter)
        self.f_data = data[:,0]*fdata_unit
        self.z_data_raw = self._ConvToCompl(data[:,1],data[:,2],dtype=dtype)

    def add_fromhdf():
        pass

    def add_froms2p(self,fname,y1_col,y2_col,dtype,fdata_unit=1.,delimiter=None):
        '''
        dtype = 'realimag', 'dBmagphaserad', 'linmagphaserad', 'dBmagphasedeg', 'linmagphasedeg'
        '''
        if dtype == 'dBmagphasedeg' or dtype == 'linmagphasedeg':
            phase_conversion = 1./180.*np.pi
        else:
            phase_conversion = 1.
        f = open(fname)
        lines = f.readlines()
        f.close()
        z_data_raw = []
        f_data = []
        if dtype=='realimag':
            for line in lines:
                if ((line!="\n") and (line[0]!="#") and (line[0]!="!")) :
                    lineinfo = line.split(delimiter)
                    f_data.append(float(lineinfo[0])*fdata_unit)
                    z_data_raw.append(np.complex(float(lineinfo[y1_col]),float(lineinfo[y2_col])))
        elif dtype=='linmagphaserad' or dtype=='linmagphasedeg':
            for line in lines:
                if ((line!="\n") and (line[0]!="#") and (line[0]!="!") and (line[0]!="M") and (line[0]!="P")):
                    lineinfo = line.split(delimiter)
                    f_data.append(float(lineinfo[0])*fdata_unit)
                    z_data_raw.append(float(lineinfo[y1_col])*np.exp( np.complex(0.,phase_conversion*float(lineinfo[y2_col]))))
        elif dtype=='dBmagphaserad' or dtype=='dBmagphasedeg':
            for line in lines:
                if ((line!="\n") and (line[0]!="#") and (line[0]!="!") and (line[0]!="M") and (line[0]!="P")):
                    lineinfo = line.split(delimiter)
                    f_data.append(float(lineinfo[0])*fdata_unit)
                    linamp = 10**(float(lineinfo[y1_col])/20.)
                    z_data_raw.append(linamp*np.exp( np.complex(0.,phase_conversion*float(lineinfo[y2_col]))))
        else:
            warnings.warn("Undefined input type! Use 'realimag', 'dBmagphaserad', 'linmagphaserad', 'dBmagphasedeg' or 'linmagphasedeg'.", SyntaxWarning)
        self.f_data = np.array(f_data)
        self.z_data_raw = np.array(z_data_raw)

    def save_fitresults(self,fname):
        pass




import numpy as np
import scipy.optimize as spopt
from scipy import stats


class circlefit(object):
    '''
    contains all the circlefit procedures
    see http://scitation.aip.org/content/aip/journal/rsi/86/2/10.1063/1.4907935
    arxiv version: http://arxiv.org/abs/1410.3365
    '''
    def _remove_cable_delay(self,f_data,z_data, delay):
        return z_data/np.exp(2j*np.pi*f_data*delay)

    def _center(self,z_data,zc):
        return z_data-zc
    
    def _dist(self,x):
        np.absolute(x,x)
        c = (x > np.pi).astype(np.int)
        return x+c*(-2.*x+2.*np.pi)  
        
    def _periodic_boundary(self,x,bound):
        return np.fmod(x,bound)-np.trunc(x/bound)*bound
        
    def _phase_fit_wslope(self,f_data,z_data,theta0, Ql, fr, slope):
        phase = np.angle(z_data)
        def residuals(p,x,y):
            theta0, Ql, fr, slope = p
            err = self._dist(y - (theta0+2.*np.arctan(2.*Ql*(1.-x/fr))-slope*x))
            return err
        p0 = [theta0, Ql, fr, slope]
        p_final = spopt.leastsq(residuals,p0,args=(np.array(f_data),np.array(phase)))
        return p_final[0]
    
    def _phase_fit(self,f_data,z_data,theta0, Ql, fr):
        phase = np.angle(z_data)
        def residuals_1(p,x,y,Ql):
            theta0, fr = p
            err = self._dist(y - (theta0+2.*np.arctan(2.*Ql*(1.-x/fr))))
            return err
        def residuals_2(p,x,y,theta0):
            Ql, fr = p
            err = self._dist(y - (theta0+2.*np.arctan(2.*Ql*(1.-x/fr))))
            return err
        def residuals_3(p,x,y,theta0,Ql):
            fr = p
            err = self._dist(y - (theta0+2.*np.arctan(2.*Ql*(1.-x/fr))))
            return err
        def residuals_4(p,x,y,theta0,fr):
            Ql = p
            err = self._dist(y - (theta0+2.*np.arctan(2.*Ql*(1.-x/fr))))
            return err
        def residuals_5(p,x,y):
            theta0, Ql, fr = p
            err = self._dist(y - (theta0+2.*np.arctan(2.*Ql*(1.-x/fr))))
            return err
        p0 = [theta0, fr]
        p_final = spopt.leastsq(lambda a,b,c: residuals_1(a,b,c,Ql),p0,args=(f_data,phase))#,ftol=1e-12,xtol=1e-12)
        theta0, fr = p_final[0]
        p0 = [Ql, fr]
        p_final = spopt.leastsq(lambda a,b,c: residuals_2(a,b,c,theta0),p0,args=(f_data,phase))#,ftol=1e-12,xtol=1e-12)
        Ql, fr = p_final[0]
        p0 = fr
        p_final = spopt.leastsq(lambda a,b,c: residuals_3(a,b,c,theta0,Ql),p0,args=(f_data,phase))#,ftol=1e-12,xtol=1e-12)
        fr = p_final[0]
        p0 = Ql
        p_final = spopt.leastsq(lambda a,b,c: residuals_4(a,b,c,theta0,fr),p0,args=(f_data,phase))#,ftol=1e-12,xtol=1e-12)
        Ql = p_final[0]
        p0 = [theta0, Ql, fr]
        p_final = spopt.leastsq(residuals_5,p0,args=(f_data,phase))
        return p_final[0]
    
    def _fit_skewed_lorentzian(self,f_data,z_data):
        amplitude = np.absolute(z_data)
        amplitude_sqr = amplitude**2
        A1a = np.minimum(amplitude_sqr[0],amplitude_sqr[-1])
        A3a = -np.max(amplitude_sqr)
        fra = f_data[np.argmin(amplitude_sqr)]
        def residuals(p,x,y):
            A2, A4, Ql = p
            err = y -(A1a+A2*(x-fra)+(A3a+A4*(x-fra))/(1.+4.*Ql**2*((x-fra)/fra)**2))
            return err
        p0 = [0., 0., 1e3]
        p_final = spopt.leastsq(residuals,p0,args=(np.array(f_data),np.array(amplitude_sqr)))
        A2a, A4a, Qla = p_final[0]
    
        def residuals2(p,x,y):
            A1, A2, A3, A4, fr, Ql = p
            err = y -(A1+A2*(x-fr)+(A3+A4*(x-fr))/(1.+4.*Ql**2*((x-fr)/fr)**2))
            return err
        def fitfunc(x,A1, A2, A3, A4, fr, Ql):
            return A1+A2*(x-fr)+(A3+A4*(x-fr))/(1.+4.*Ql**2*((x-fr)/fr)**2)
            
        p0 = [A1a, A2a , A3a, A4a, fra, Qla]
        #p_final = spopt.leastsq(residuals2,p0,args=(np.array(f_data),np.array(amplitude_sqr)))
        try:
            popt, pcov = spopt.curve_fit(fitfunc, np.array(f_data), np.array(amplitude_sqr),p0=p0)
        #A1, A2, A3, A4, fr, Ql = p_final[0]
        #print p_final[0][5]
            if pcov is not None:
                self.df_error = np.sqrt(pcov[4][4])
                self.dQl_error = np.sqrt(pcov[5][5])
            else:
                self.df_error = np.inf
                self.dQl_error = np.inf
        except:
            popt = p0
            self.df_error = np.inf
            self.dQl_error = np.inf
        #return p_final[0]
        return popt
    
    def _fit_circle(self,z_data, refine_results=False):
        def calc_moments(z_data):
            xi = z_data.real
            xi_sqr = xi*xi
            yi = z_data.imag
            yi_sqr = yi*yi
            zi = xi_sqr+yi_sqr
            Nd = float(len(xi))
            xi_sum = xi.sum()
            yi_sum = yi.sum()
            zi_sum = zi.sum()
            xiyi_sum = (xi*yi).sum()
            xizi_sum = (xi*zi).sum()
            yizi_sum = (yi*zi).sum()
            return np.array([ [(zi*zi).sum(), xizi_sum, yizi_sum, zi_sum],  \
            [xizi_sum, xi_sqr.sum(), xiyi_sum, xi_sum], \
            [yizi_sum, xiyi_sum, yi_sqr.sum(), yi_sum], \
            [zi_sum, xi_sum, yi_sum, Nd] ])
    
        M = calc_moments(z_data)
    
        a0 = ((M[2][0]*M[3][2]-M[2][2]*M[3][0])*M[1][1]-M[1][2]*M[2][0]*M[3][1]-M[1][0]*M[2][1]*M[3][2]+M[1][0]*M[2][2]*M[3][1]+M[1][2]*M[2][1]*M[3][0])*M[0][3]+(M[0][2]*M[2][3]*M[3][0]-M[0][2]*M[2][0]*M[3][3]+M[0][0]*M[2][2]*M[3][3]-M[0][0]*M[2][3]*M[3][2])*M[1][1]+(M[0][1]*M[1][3]*M[3][0]-M[0][1]*M[1][0]*M[3][3]-M[0][0]*M[1][3]*M[3][1])*M[2][2]+(-M[0][1]*M[1][2]*M[2][3]-M[0][2]*M[1][3]*M[2][1])*M[3][0]+((M[2][3]*M[3][1]-M[2][1]*M[3][3])*M[1][2]+M[2][1]*M[3][2]*M[1][3])*M[0][0]+(M[1][0]*M[2][3]*M[3][2]+M[2][0]*(M[1][2]*M[3][3]-M[1][3]*M[3][2]))*M[0][1]+((M[2][1]*M[3][3]-M[2][3]*M[3][1])*M[1][0]+M[1][3]*M[2][0]*M[3][1])*M[0][2]
        a1 = (((M[3][0]-2.*M[2][2])*M[1][1]-M[1][0]*M[3][1]+M[2][2]*M[3][0]+2.*M[1][2]*M[2][1]-M[2][0]*M[3][2])*M[0][3]+(2.*M[2][0]*M[3][2]-M[0][0]*M[3][3]-2.*M[2][2]*M[3][0]+2.*M[0][2]*M[2][3])*M[1][1]+(-M[0][0]*M[3][3]+2.*M[0][1]*M[1][3]+2.*M[1][0]*M[3][1])*M[2][2]+(-M[0][1]*M[1][3]+2.*M[1][2]*M[2][1]-M[0][2]*M[2][3])*M[3][0]+(M[1][3]*M[3][1]+M[2][3]*M[3][2])*M[0][0]+(M[1][0]*M[3][3]-2.*M[1][2]*M[2][3])*M[0][1]+(M[2][0]*M[3][3]-2.*M[1][3]*M[2][1])*M[0][2]-2.*M[1][2]*M[2][0]*M[3][1]-2.*M[1][0]*M[2][1]*M[3][2])
        a2 = ((2.*M[1][1]-M[3][0]+2.*M[2][2])*M[0][3]+(2.*M[3][0]-4.*M[2][2])*M[1][1]-2.*M[2][0]*M[3][2]+2.*M[2][2]*M[3][0]+M[0][0]*M[3][3]+4.*M[1][2]*M[2][1]-2.*M[0][1]*M[1][3]-2.*M[1][0]*M[3][1]-2.*M[0][2]*M[2][3])
        a3 = (-2.*M[3][0]+4.*M[1][1]+4.*M[2][2]-2.*M[0][3])
        a4 = -4.
    
        def func(x):
            return a0+a1*x+a2*x*x+a3*x*x*x+a4*x*x*x*x
    
        def d_func(x):
            return a1+2*a2*x+3*a3*x*x+4*a4*x*x*x
    
        x0 = spopt.fsolve(func, 0., fprime=d_func)
    
        def solve_eq_sys(val,M):
            #prepare
            M[3][0] = M[3][0]+2*val
            M[0][3] = M[0][3]+2*val
            M[1][1] = M[1][1]-val
            M[2][2] = M[2][2]-val
            return np.linalg.svd(M)
    
        U,s,Vt = solve_eq_sys(x0[0],M)
    
        A_vec = Vt[np.argmin(s),:]
    
        xc = -A_vec[1]/(2.*A_vec[0])
        yc = -A_vec[2]/(2.*A_vec[0])
        # the term *sqrt term corrects for the constraint, because it may be altered due to numerical inaccuracies during calculation
        r0 = 1./(2.*np.absolute(A_vec[0]))*np.sqrt(A_vec[1]*A_vec[1]+A_vec[2]*A_vec[2]-4.*A_vec[0]*A_vec[3])
        if refine_results:
            print("agebraic r0: " + str(r0))
            xc,yc,r0 = self._fit_circle_iter(z_data, xc, yc, r0)
            r0 = self._fit_circle_iter_radialweight(z_data, xc, yc, r0)
            print("iterative r0: " + str(r0))
        return xc, yc, r0

    def _guess_delay(self,f_data,z_data):
        phase2 = np.unwrap(np.angle(z_data))
        gradient, intercept, r_value, p_value, std_err = stats.linregress(f_data,phase2)
        return gradient*(-1.)/(np.pi*2.)
    
    
    def _fit_delay(self,f_data,z_data,delay=0.,maxiter=0):
        def residuals(p,x,y):
            phasedelay = p
            z_data_temp = y*np.exp(1j*(2.*np.pi*phasedelay*x))
            xc,yc,r0 = self._fit_circle(z_data_temp)
            err = np.sqrt((z_data_temp.real-xc)**2+(z_data_temp.imag-yc)**2)-r0
            return err
        p_final = spopt.leastsq(residuals,delay,args=(f_data,z_data),maxfev=maxiter,ftol=1e-12,xtol=1e-12)
        return p_final[0][0]
    
    def _fit_delay_alt_bigdata(self,f_data,z_data,delay=0.,maxiter=0):
        def residuals(p,x,y):
            phasedelay = p
            z_data_temp = 1j*2.*np.pi*phasedelay*x
            np.exp(z_data_temp,out=z_data_temp)
            np.multiply(y,z_data_temp,out=z_data_temp)
            #z_data_temp = y*np.exp(1j*(2.*np.pi*phasedelay*x))
            xc,yc,r0 = self._fit_circle(z_data_temp)
            err = np.sqrt((z_data_temp.real-xc)**2+(z_data_temp.imag-yc)**2)-r0
            return err
        p_final = spopt.leastsq(residuals,delay,args=(f_data,z_data),maxfev=maxiter,ftol=1e-12,xtol=1e-12)
        return p_final[0][0]
    
    def _fit_entire_model(self,f_data,z_data,fr,absQc,Ql,phi0,delay,a=1.,alpha=0.,maxiter=0):
        '''
        fits the whole model: a*exp(i*alpha)*exp(-2*pi*i*f*delay) * [ 1 - {Ql/Qc*exp(i*phi0)} / {1+2*i*Ql*(f-fr)/fr} ]
        '''
        def funcsqr(p,x):
            fr,absQc,Ql,phi0,delay,a,alpha = p
            return np.array([np.absolute( ( a*np.exp(np.complex(0,alpha))*np.exp(np.complex(0,-2.*np.pi*delay*x[i])) * ( 1 - (Ql/absQc*np.exp(np.complex(0,phi0)))/(np.complex(1,2*Ql*(x[i]-fr)/fr)) )  ) )**2 for i in range(len(x))])
        def residuals(p,x,y):
            fr,absQc,Ql,phi0,delay,a,alpha = p
            err = [np.absolute( y[i] - ( a*np.exp(np.complex(0,alpha))*np.exp(np.complex(0,-2.*np.pi*delay*x[i])) * ( 1 - (Ql/absQc*np.exp(np.complex(0,phi0)))/(np.complex(1,2*Ql*(x[i]-fr)/fr)) )  ) ) for i in range(len(x))]
            return err
        p0 = [fr,absQc,Ql,phi0,delay,a,alpha]
        (popt, params_cov, infodict, errmsg, ier) = spopt.leastsq(residuals,p0,args=(np.array(f_data),np.array(z_data)),full_output=True,maxfev=maxiter)
        len_ydata = len(np.array(f_data))
        if (len_ydata > len(p0)) and params_cov is not None:  #p_final[1] is cov_x data  #this caculation is from scipy curve_fit routine - no idea if this works correctly...
            s_sq = (funcsqr(popt, np.array(f_data))).sum()/(len_ydata-len(p0))
            params_cov = params_cov * s_sq
        else:
            params_cov = np.inf
        return popt, params_cov, infodict, errmsg, ier
    
    #
    
    def _optimizedelay(self,f_data,z_data,Ql,fr,maxiter=4):
        xc,yc,r0 = self._fit_circle(z_data)
        z_data = self._center(z_data,np.complex(xc,yc))
        theta, Ql, fr, slope = self._phase_fit_wslope(f_data,z_data,0.,Ql,fr,0.)
        delay = 0.
        for i in range(maxiter-1): #interate to get besser phase delay term
            delay = delay - slope/(2.*2.*np.pi)
            z_data_corr = self._remove_cable_delay(f_data,z_data,delay)
            xc, yc, r0 = self._fit_circle(z_data_corr)
            z_data_corr2 = self._center(z_data_corr,np.complex(xc,yc))
            theta0, Ql, fr, slope = self._phase_fit_wslope(f_data,z_data_corr2,0.,Ql,fr,0.)
        delay = delay - slope/(2.*2.*np.pi)  #start final interation
        return delay
    
    def _fit_circle_iter(self,z_data, xc, yc, rc):
        '''
        this is the radial weighting procedure
        it improves your fitting value for the radius = Ql/Qc
        use this to improve your fit in presence of heavy noise
        after having used the standard algebraic fir_circle() function
        the weight here is: W=1/sqrt((xc-xi)^2+(yc-yi)^2)
        this works, because the center of the circle is usually much less
        corrupted by noise than the radius
        '''
        xdat = z_data.real
        ydat = z_data.imag
        def fitfunc(x,y,xc,yc):
            return np.sqrt((x-xc)**2+(y-yc)**2)
        def residuals(p,x,y):
            xc,yc,r = p
            temp = (r-fitfunc(x,y,xc,yc))
            return temp
        p0 = [xc,yc,rc]
        p_final = spopt.leastsq(residuals,p0,args=(xdat,ydat))
        xc,yc,rc = p_final[0]
        return xc,yc,rc
    
    def _fit_circle_iter_radialweight(self,z_data, xc, yc, rc):
        '''
        this is the radial weighting procedure
        it improves your fitting value for the radius = Ql/Qc
        use this to improve your fit in presence of heavy noise
        after having used the standard algebraic fir_circle() function
        the weight here is: W=1/sqrt((xc-xi)^2+(yc-yi)^2)
        this works, because the center of the circle is usually much less
        corrupted by noise than the radius
        '''
        xdat = z_data.real
        ydat = z_data.imag
        def fitfunc(x,y):
            return np.sqrt((x-xc)**2+(y-yc)**2)
        def weight(x,y):
            try:
                res = 1./np.sqrt((xc-x)**2+(yc-y)**2)
            except:
                res = 1.
            return res
        def residuals(p,x,y):
            r = p[0]
            temp = (r-fitfunc(x,y))*weight(x,y)
            return temp
        p0 = [rc]
        p_final = spopt.leastsq(residuals,p0,args=(xdat,ydat))
        return p_final[0][0]
    
    def _get_errors(self,residual,xdata,ydata,fitparams):
        '''
        wrapper for get_cov, only gives the errors and chisquare
        '''
        chisqr, cov = self._get_cov(residual,xdata,ydata,fitparams)
        if cov is not None:
            errors = np.sqrt(np.diagonal(cov))
        else:
            errors = None
        return chisqr, errors
    
    def _residuals_notch_full(self,p,x,y):
        fr,absQc,Ql,phi0,delay,a,alpha = p
        err = np.absolute( y - ( a*np.exp(np.complex(0,alpha))*np.exp(np.complex(0,-2.*np.pi*delay*x)) * ( 1 - (Ql/absQc*np.exp(np.complex(0,phi0)))/(np.complex(1,2*Ql*(x-fr)/float(fr))) )  ) )
        return err
    
    def _residuals_notch_ideal(self,p,x,y):
        fr,absQc,Ql,phi0 = p
        #if fr == 0: print p
        err = np.absolute( y - (  ( 1. - (Ql/float(absQc)*np.exp(1j*phi0))/(1+2j*Ql*(x-fr)/float(fr)) )  ) )
        #if np.isinf((np.complex(1,2*Ql*(x-fr)/float(fr))).imag):
         #   print np.complex(1,2*Ql*(x-fr)/float(fr))
          #  print "x: " + str(x)
           # print "Ql: " +str(Ql)
            #print "fr: " +str(fr)
        return err
    
    def _residuals_notch_ideal_complex(self,p,x,y):
        fr,absQc,Ql,phi0 = p
        #if fr == 0: print p
        err = y - (  ( 1. - (Ql/float(absQc)*np.exp(1j*phi0))/(1+2j*Ql*(x-fr)/float(fr)) )  )
        #if np.isinf((np.complex(1,2*Ql*(x-fr)/float(fr))).imag):
         #   print np.complex(1,2*Ql*(x-fr)/float(fr))
          #  print "x: " + str(x)
           # print "Ql: " +str(Ql)
            #print "fr: " +str(fr)
        return err
        
    def _residuals_directrefl(self,p,x,y):
        fr,Qc,Ql = p
        #if fr == 0: print p
        err = y - ( 2.*Ql/Qc - 1. + 2j*Ql*(fr-x)/fr ) / ( 1. - 2j*Ql*(fr-x)/fr )
        #if np.isinf((np.complex(1,2*Ql*(x-fr)/float(fr))).imag):
         #   print np.complex(1,2*Ql*(x-fr)/float(fr))
          #  print "x: " + str(x)
           # print "Ql: " +str(Ql)
            #print "fr: " +str(fr)
        return err
    
    def _residuals_transm_ideal(self,p,x,y):
        fr,Ql = p
        err = np.absolute( y -   ( 1./(np.complex(1,2*Ql*(x-fr)/float(fr))) )   )
        return err
    
    
    def _get_cov_fast_notch(self,xdata,ydata,fitparams): #enhanced by analytical derivatives
        #derivatives of notch_ideal model with respect to parameters
        def dS21_dQl(p,f):
            fr,absQc,Ql,phi0 = p
            return - (np.exp(1j*phi0) * fr**2) / (absQc * (fr+2j*Ql*f-2j*Ql*fr)**2 )
    
        def dS21_dQc(p,f):
            fr,absQc,Ql,phi0 = p
            return (np.exp(1j*phi0) * Ql*fr) / (2j*(f-fr)*absQc**2*Ql+absQc**2*fr )
    
        def dS21_dphi0(p,f):
            fr,absQc,Ql,phi0 = p
            return - (1j*Ql*fr*np.exp(1j*phi0) ) / (2j*(f-fr)*absQc*Ql+absQc*fr )
    
        def dS21_dfr(p,f):
            fr,absQc,Ql,phi0 = p
            return - (2j*Ql**2*f*np.exp(1j*phi0) ) / (absQc * (fr+2j*Ql*f-2j*Ql*fr)**2 )
    
        u = self._residuals_notch_ideal_complex(fitparams,xdata,ydata)
        chi = np.absolute(u)
        u = u/chi  # unit vector pointing in the correct direction for the derivative
    
        aa = dS21_dfr(fitparams,xdata)
        bb = dS21_dQc(fitparams,xdata)
        cc = dS21_dQl(fitparams,xdata)
        dd = dS21_dphi0(fitparams,xdata)
    
        Jt = np.array([aa.real*u.real+aa.imag*u.imag, bb.real*u.real+bb.imag*u.imag\
                , cc.real*u.real+cc.imag*u.imag, dd.real*u.real+dd.imag*u.imag  ])
        A = np.dot(Jt,np.transpose(Jt))
        chisqr = 1./float(len(xdata)-len(fitparams)) * (chi**2).sum()
        try:
            cov = np.linalg.inv(A)*chisqr
        except:
            cov = None
        return chisqr, cov
        
    def _get_cov_fast_directrefl(self,xdata,ydata,fitparams): #enhanced by analytical derivatives
        #derivatives of notch_ideal model with respect to parameters
        def dS21_dQl(p,f):
            fr,Qc,Ql = p
            return 2.*fr**2/( Qc*(2j*Ql*fr-2j*Ql*f+fr)**2 )
    
        def dS21_dQc(p,f):
            fr,Qc,Ql = p
            return 2.*Ql*fr/(2j*Qc**2*(f-fr)*Ql-Qc**2*fr)
    
        def dS21_dfr(p,f):
            fr,Qc,Ql = p
            return - 4j*Ql**2*f/(Qc*(2j*Ql*fr-2j*Ql*f+fr)**2)
    
        u = self._residuals_directrefl(fitparams,xdata,ydata)
        chi = np.absolute(u)
        u = u/chi  # unit vector pointing in the correct direction for the derivative
    
        aa = dS21_dfr(fitparams,xdata)
        bb = dS21_dQc(fitparams,xdata)
        cc = dS21_dQl(fitparams,xdata)
    
        Jt = np.array([aa.real*u.real+aa.imag*u.imag, bb.real*u.real+bb.imag*u.imag\
                , cc.real*u.real+cc.imag*u.imag  ])
        A = np.dot(Jt,np.transpose(Jt))
        chisqr = 1./float(len(xdata)-len(fitparams)) * (chi**2).sum()
        try:
            cov = np.linalg.inv(A)*chisqr
        except:
            cov = None
        return chisqr, cov



import numpy as np
from scipy import sparse
from scipy.interpolate import interp1d

class calibration(object):
    '''
    some useful tools for manual calibration
    '''
    def normalize_zdata(self,z_data,cal_z_data):
        return z_data/cal_z_data
        
    def normalize_amplitude(self,z_data,cal_ampdata):
        return z_data/cal_ampdata
        
    def normalize_phase(self,z_data,cal_phase):
        return z_data*np.exp(-1j*cal_phase)
        
    def normalize_by_func(self,f_data,z_data,func):
        return z_data/func(f_data)
        
    def _baseline_als(self,y, lam, p, niter=10):
        '''
        see http://zanran_storage.s3.amazonaws.com/www.science.uva.nl/ContentPages/443199618.pdf
        "Asymmetric Least Squares Smoothing" by P. Eilers and H. Boelens in 2005.
        http://stackoverflow.com/questions/29156532/python-baseline-correction-library
        "There are two parameters: p for asymmetry and lambda for smoothness. Both have to be
        tuned to the data at hand. We found that generally 0.001<=p<=0.1 is a good choice
        (for a signal with positive peaks) and 10e2<=lambda<=10e9, but exceptions may occur."
        '''
        L = len(y)
        D = sparse.csc_matrix(np.diff(np.eye(L), 2))
        w = np.ones(L)
        for i in range(niter):
            W = sparse.spdiags(w, 0, L, L)
            Z = W + lam * D.dot(D.transpose())
            z = sparse.linalg.spsolve(Z, w*y)
            w = p * (y > z) + (1-p) * (y < z)
        return z
        
    def fit_baseline_amp(self,z_data,lam,p,niter=10):
        '''
        for this to work, you need to analyze a large part of the baseline
        tune lam and p until you get the desired result
        '''
        return self._baseline_als(np.absolute(z_data),lam,p,niter=niter)
    
    def baseline_func_amp(self,z_data,f_data,lam,p,niter=10):
        '''
        for this to work, you need to analyze a large part of the baseline
        tune lam and p until you get the desired result
        returns the baseline as a function
        the points in between the datapoints are computed by cubic interpolation
        '''
        return interp1d(f_data, self._baseline_als(np.absolute(z_data),lam,p,niter=niter), kind='cubic')
        
    def baseline_func_phase(self,z_data,f_data,lam,p,niter=10):
        '''
        for this to work, you need to analyze a large part of the baseline
        tune lam and p until you get the desired result
        returns the baseline as a function
        the points in between the datapoints are computed by cubic interpolation
        '''
        return interp1d(f_data, self._baseline_als(np.angle(z_data),lam,p,niter=niter), kind='cubic')
        
    def fit_baseline_phase(self,z_data,lam,p,niter=10):
        '''
        for this to work, you need to analyze a large part of the baseline
        tune lam and p until you get the desired result
        '''
        return self._baseline_als(np.angle(z_data),lam,p,niter=niter)



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

config = {
    'description': 'resonator_tools fork from vdrhtc',
    'author': 'Sebastian Probst',
    'url': 'https://github.com/vdrhtc/resonator_tools',
    'download_url': 'https://github.com/vdrhtc/resonator_tools',
    'author_email': 'seb.probst@gmail.com',
    'version': '0.1',
    'install_requires': ['nose'],
    'packages': ['resonator_tools'],
    'scripts': [],
    'name': 'resonator_tools_vdrhtc'
}

setup(**config)
